#include "HL1606stripPWM.h"
// Written from scratch, but inspired by 
// http://waitingforbigo.com/2010/10/02/second-preview-release-of-fastspi_led-library
// This library can PWM an HL1606 strip with 3 or 4 bits of color per LED
// See the example sketches for more detailed usage

// (c) Adafruit Industries / Limor Fried 2010. Released under MIT license.

#define DEBUG(d) if(SerialUSB.isConnected() && (SerialUSB.getDTR() || SerialUSB.getRTS())) { SerialUSB.print(millis(), DEC); SerialUSB.print(" - "); SerialUSB.println(d); }
#define DEBUG_VAL(d, v, t) if(SerialUSB.isConnected() && (SerialUSB.getDTR() || SerialUSB.getRTS())) { SerialUSB.print(millis(), DEC); SerialUSB.print(" - "); SerialUSB.print(d); SerialUSB.println(v, t); }

static uint8_t pwmincr;
static uint8_t pwmcounter;

// the arrays of bytes that hold each LED's PWM values
static uint8_t *redPWM;
static uint8_t *greenPWM;
static uint8_t *bluePWM;

// how many LEDs
static uint8_t nLEDs;

static uint8_t latchPin;



//pointer to spi object
static HardwareSPI *SPIn;

//pointer to timer object
static HardwareTimer *timer;


// functions for hardware SPI
/** Send a byte to the card */
static void spiSend(uint8_t b) 
{
	SPIn->send(b);
	//while(SPIn->busy())
	//{
	//}
}
/** Receive a byte from the card */
static  uint8_t spiRec(void) 
{
	return SPIn->send(0XFF);
//	return SPIn->read();
}

HL1606stripPWM::HL1606stripPWM(uint8_t n, HardwareSPI *s, HardwareTimer *t, uint8_t l) {
  nLEDs = n;
	SPIn = s;
	timer = t;
	latchPin = l;


  pwmcounter = 0;
  PWMbits = 3;
  pwmincr = 256 / (1 << PWMbits);
  SPIspeed = SPI_562_500KHZ;

  redPWM = (uint8_t *)malloc(nLEDs);
  greenPWM = (uint8_t *)malloc(nLEDs);
  bluePWM = (uint8_t *)malloc(nLEDs);
  for (uint8_t i=0; i< nLEDs; i++) {
    setLEDcolorPWM(i, 0, 0, 0);
  }
}

void HL1606stripPWM::begin(void) {
	pinMode(latchPin, OUTPUT);

  SPIinit();
  timerinit();

}


void timerTick() {
  uint8_t i, d;
  // write out data to strip 
  for (i=0; i< nLEDs; i++) {
    d = 0x80;          // set the latch bit
    // calculate the next LED's byte
    if (pwmcounter < redPWM[i]) {
      d |= 0x04;
    } 
    if (pwmcounter < greenPWM[i]) {
      d |= 0x10;
    } 
    if (pwmcounter < bluePWM[i]) {
      d |= 0x01;
    } 

    // check that previous xfer completed
    //while(!(SPSR & _BV(SPIF))); 
   //  while (!spi_is_tx_empty(SPIn->spi_d));
 
    // send new data
    //SPDR = d; 
    spiSend(d);
  }
  //DEBUG_VAL("timerTick() 0x", d, HEX);
  //DEBUG_VAL("pwmcounter 0x", pwmcounter, HEX);
  //DEBUG_VAL("pwmincr 0x", pwmincr, HEX);

  // increment our PWM counter
  pwmcounter += pwmincr;

  // make sure we're all done
  //while(!(SPSR & _BV(SPIF)));
 
 // while (!spi_is_tx_empty(SPIn->spi_d));

  // latch
  digitalWrite(latchPin, HIGH);
  delayMicroseconds(3);
  digitalWrite(latchPin, LOW);
}


void HL1606stripPWM::timerinit(void) {
/*  // calculate how long it will take to pulse one strip down
  double time = nLEDs;    // each LED
  time *= 8;              // 8 bits of data for each LED;
  time *= SPIspeedDiv;    // larger dividers = more time per bit
  time *= 1000;           // time in milliseconds
  time /= F_CPU;          // multiplied by how long it takes for one instruction (nverse of cpu)
  
  time *= 100;            // calculate what percentage of CPU we can use (multiply by time)
  time /= CPUmaxpercent;
*/
  
  timer->pause();
  timer->setPeriod(1 * 1000);	// in microseconds
	// Set up an interrupt on channel 1
	timer->setMode(TIMER_CH1, TIMER_OUTPUT_COMPARE);
	timer->setCompare(TIMER_CH1, 1);  // Interrupt 1 count after each update
	timer->attachInterrupt(TIMER_CH1, timerTick);


  
	timer->refresh();
	timer->resume();

  
  // set up the interrupt to write to the entire strip
  // Each pixel requires a 1-byte SPI transfer, so with a n-pixel strip, thats n bytes at 1 MHz
  // which makes for a 1 MHz / n PWM frequency. Say for a 100 LED strip, we can update no faster
  // than 10 KHz. Lets make it 1 KHz to start

} 



void HL1606stripPWM::setLEDcolorPWM(uint8_t n, uint8_t r, uint8_t g, uint8_t b) {
     redPWM[n] = r; 
     greenPWM[n] = g; 
     bluePWM[n] = b;
}





void HL1606stripPWM::SPIinit(void) {
  
	// Turn on the SPI port
	SPIn->begin(SPI_562_500KHZ, MSBFIRST, 0);

  
  /*pinMode(DATA_PIN, OUTPUT);
  pinMode(CLOCK_PIN, OUTPUT);
  pinMode(latchPin, OUTPUT);
  
  // set up high speed SPI for 500 KHz
  // The datasheet says that the clock pulse width must be > 300ns. Two pulses > 600ns that would
  // make the max frequency 1.6 MHz - fat chance getting that out of HL1606's though
  SPCR = _BV(SPE) | _BV(MSTR);   // enable SPI master mode
  setSPIdivider(SPIspeedDiv);          // SPI clock is FCPU/32 = 500 Khz for most arduinos
  
  // we send a fake SPI byte to get the 'finished' bit set in the register, dont remove!!!
  SPDR = 0;*/
  spiSend(0);
}

uint8_t HL1606stripPWM::getSPIspeed(void) { return SPIspeed; }

void HL1606stripPWM::setSPIspeed(uint8_t spispeed) {
  SPIspeed = spispeed;

	SPIinit();
  timerinit();
}

uint8_t HL1606stripPWM::numLEDs(void) {
  return nLEDs;
}

void HL1606stripPWM::setPWMbits(uint8_t b) {
  PWMbits = b;
  pwmincr = 256 / (1 << PWMbits);
}

uint8_t HL1606stripPWM::getPWMbits(void) {
  return PWMbits;
}


void HL1606stripPWM::enable(void) {
	timer->resume();
}

void HL1606stripPWM::disable(void) {
	timer->pause();
}